from abc import abstractmethod, ABC
from typing import List, Generator
from Shared.NeumDocument import NeumDocument
from Shared.LocalFile import LocalFile
from Loaders.Loader import Loader
from langchain.document_loaders import PyPDFLoader

class PDFLoader(Loader):
    """" PDF Loader """
    """" loader_information contains: """
    """ [] """

    @property
    def loader_name(self) -> str:
        return "PDFLoader"
    
    @property
    def requiredProperties(self) -> List[str]:
        return []

    @property
    def optionalProperties(self) -> List[str]:
        return []

    @property
    def availableMetadata(self) -> List[str]:
        return []

    @property
    def availableContent(self) -> List[str]:
        return []

    #Probably worth re-writing directly on top of pypdf to get access
    #to more metadata including images, tables, etc.
    def load(self, file:LocalFile) -> Generator[NeumDocument, None, None]:
        """Load data into Document objects."""
        loader = PyPDFLoader(file_path=file.file_path)
        documents = loader.load()
        # join the file and document metadata objects
        for doc in documents:
            doc.metadata.update(file.metadata)
            yield NeumDocument(id=file.id, content=doc.page_content, metadata=doc.metadata)
    
    def validate(self) -> bool:
        return True   


from abc import abstractmethod, ABC
from typing import List, Generator
from Shared.NeumDocument import NeumDocument
from Shared.LocalFile import LocalFile
from Loaders.Loader import Loader
from neumai_tools import JSONLoader

class NeumJSONLoader(Loader):
    """" Neum Website Connector """

    @property
    def requiredProperties(self) -> List[str]:
        return []

    @property
    def optionalProperties(self) -> List[str]:
        return ["id_key"]

    @property
    def loader_name(self) -> str:
        return "NeumJSONLoader"
    
    @property
    def availableMetadata(self) -> List[str]:
        return ["custom"]

    @property
    def availableContent(self) -> List[str]:
        return ["custom"]

    def load(self, file:LocalFile) -> Generator[NeumDocument, None, None]:
        """Load data into Document objects."""
        id_key = self.loader_information.get('id_key', "id")
        if file.file_path != None:
            loader = JSONLoader(file_path=file.file_path, id_key=id_key, selector=self.selector)
        elif file.in_mem_data != None:
            loader = JSONLoader(file_data=file.in_mem_data, id_key=id_key, selector=self.selector)
        
        documents = loader.load()
        # join the file and document metadata objects
        for doc in documents:
            doc.metadata.update(file.metadata)
            yield doc

    def validate(self) -> bool:
        return True   


from psycopg2.extras import DictCursor
from datetime import datetime
from DataConnector import DataConnector
from typing import List, Generator
from Shared.LocalFile import LocalFile
from Shared.CloudFile import CloudFile
import singlestoredb as s2
import json


class SingleStoreConnector(DataConnector):
    """SingleStore Connector \n
    connector_information requires:\n
    [ url, query ]"""
    
    @property
    def connector_name(self) -> str:
        return "SingleStoreConnector"
    
    @property
    def requiredProperties(self) -> List[str]:
        return ["url", "query"]

    @property
    def optionalProperties(self) -> List[str]:
        return ["batch_size"]
    
    @property
    def availableMetadata(self) -> str:
        return []

    @property
    def availableContent(self) -> str:
        return ['row']
    
    @property
    def schedule_avaialable(self) -> bool:
        return True

    @property
    def auto_sync_available(self) -> bool:
        return False
    
    @property
    def compatible_loaders(self) -> List[str]:
        return ["NeumJSONLoader"]
    
    def datetime_serializer(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        raise TypeError("Type not serializable")

    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        url = self.connector_information['url']
        query = self.connector_information['query']
        batch_size = 1000
        if 'batch_size' in self.connector_information.keys():
            batch_size = self.connector_information['batch_size']
        
        with s2.connect(url, results_type="dict") as conn:
            with conn.cursor() as cur:
                batch_rows = []
                cur.execute(query)
                while True:
                    rows = cur.fetchmany(batch_size)
                    if not rows:
                        break

                    for row in rows:
                        batch_rows.append(row)
                        if(len(batch_rows) == batch_size):
                            json_object = json.dumps(batch_rows, default=self.datetime_serializer)
                            yield CloudFile(data=json_object, metadata={})
                            batch_rows = []

                if len(batch_rows) > 0:
                        json_object = json.dumps(batch_rows, default=self.datetime_serializer)
                        yield CloudFile(data=json_object, metadata={})

    def connect_and_list_delta(self, last_run:datetime) -> Generator[CloudFile, None, None]:
        # No metadatadata to determine what rows are new. Needs to be done through websocket
        yield from self.connect_and_list_full()

    def connect_and_download(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
         data = json.loads(cloudFile.data)
         for row in data:
            yield LocalFile(in_mem_data=dict(row), metadata=cloudFile.metadata)
    
    def validate(self) -> bool:
        try:
            url = self.connector_information['url']
            query = self.connector_information['query']
        except:
            raise ValueError("Required properties not set")
        
        try: 
            s2.connect(url, results_type="dict")
        except:
            raise Exception("Connection URL is incorrect")
        return True 
                


from abc import abstractmethod, ABC
from typing import List, Generator
from Shared.LocalFile import LocalFile
from Shared.CloudFile import CloudFile
from Shared.Selector import Selector
from Loaders.Loader import Loader
from datetime import datetime

class DataConnector(ABC):
    def __init__(self, connector_information:dict = {}, selector:Selector = Selector(to_embed=[], to_metadata=[])) -> None:
        self.connector_information = connector_information
        self.selector = selector

    @property
    @abstractmethod
    def connector_name(self) -> str:
        pass

    @property
    @abstractmethod
    def requiredProperties(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def optionalProperties(self) -> List[str]:
        pass
    
    @property
    @abstractmethod
    def availableMetadata(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def availableContent(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def schedule_avaialable(self) -> bool:
        pass

    @property
    @abstractmethod
    def auto_sync_available(self) -> bool:
        pass

    @property
    @abstractmethod
    def compatible_loaders(self) -> List[str]:
        pass

    @abstractmethod
    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        """Connect to source and download file into local storage"""
    
    @abstractmethod
    def connect_and_list_delta(self, last_run:datetime) -> Generator[LocalFile, None, None]:
        """Check for changes in the source"""
        """Code to be pushed to a worker and run on a schedule"""
    
    @abstractmethod
    def connect_and_download(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        """Connect to source and download file into local storage"""

    @abstractmethod
    def validate(self) -> bool:
        """Validate if the connector is correctly configured"""

    # To do auto_sync logic.

    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['connector_name'] = self.connector_name
        json_to_return['connector_information'] = self.connector_information
        json_to_return['selector'] = self.selector.toJson()
        return json_to_return
    
    def to_model(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..
        This is different than toJson, here we use it to create a model, we don't want to return the api key in the body back. Eventualyl this should be its own class...
        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['connector_name'] = self.connector_name
        json_to_return['connector_information'] = self.connector_information
        json_to_return['selector'] = self.selector.to_model()
        return json_to_return

    def config(self):
        json_to_return = {}
        json_to_return['requiredProperties'] = self.requiredProperties
        json_to_return['optionalProperties'] = self.optionalProperties
        json_to_return['availableMetadata'] = self.availableMetadata
        json_to_return['availableContent'] = self.availableContent
        json_to_return['compatible_loaders'] = self.compatible_loaders
        json_to_return['schedule_avaialable'] = self.schedule_avaialable
        json_to_return['auto_sync_available'] = self.auto_sync_available
        return json_to_return


from abc import abstractmethod, ABC
from typing import List, Generator
from Shared.NeumDocument import NeumDocument
from Shared.LocalFile import LocalFile
from Loaders.Loader import Loader
from neumai_tools import CSVLoader

class NeumCSVLoader(Loader):
    """" Neum Website Connector """

    @property
    def loader_name(self) -> str:
        return "NeumCSVLoader"
    
    @property
    def requiredProperties(self) -> List[str]:
        return []

    @property
    def optionalProperties(self) -> List[str]:
        return ["id_key" , "source_column" , "encoding"]
    
    @property
    def availableMetadata(self) -> List[str]:
        return ["custom"]

    @property
    def availableContent(self) -> List[str]:
        return ["custom"]

    def load(self, file:LocalFile) -> Generator[NeumDocument, None, None]:
        """Load data into Document objects."""
        source_column = self.loader_information.get('source_coulmn', None)
        encoding = self.loader_information.get('encoding', "utf-8") # modify to use encoding
        id_key = self.loader_information.get('id_key', 'id') # default to id
        selector = self.selector
        embed_keys = selector.to_embed
        metadata_keys = selector.to_metadata
        try:
            loader = CSVLoader(file_path=file.file_path, id_key=id_key, source_column=source_column, embed_keys=embed_keys, metadata_keys=metadata_keys, encoding=encoding)
            documents = loader.load()
        except UnicodeDecodeError as e:
            loader = CSVLoader(file_path=file.file_path, id_key=id_key, source_column=source_column, embed_keys=embed_keys, metadata_keys=metadata_keys, encoding="cp1252-8")
            documents = loader.load()
        for doc in documents:
            doc.metadata.update(file.metadata)
            yield doc

    def validate(self) -> bool:
        return True   